/*
    This file is part of "now", a simple calendar tool
    Copyright (C) 2012-2019  Laurent Poirrier

    "now" is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include "bio.h"
#include "memory.h"
#include "config.h"

#define DEBUG(...)  fprintf(stderr, __VA_ARGS__)

// **************************************************************************
// 
// **************************************************************************
size_t unescape(char *str, size_t len)
{
	size_t t, o;
	int v;
	char c;
	
	o = 0;
	for (t = 0; t < len; t++) {
		c = str[t];
		
		if (c != '\\') {
			str[o] = c;
			o++;
			continue;
		}
		
		if ((t + 1) >= len) {
			str[o] = '\\';
			o++;
			break;
		}
		
		t++;
		c = str[t];
		if (c == 'n') {
			str[o] = '\n';
			o++;
		} else if (c == 'r') {
			str[o] = '\r';
			o++;
		} else if (c == 't') {
			str[o] = '\t';
			o++;
		} else if (c == 'a') {
			str[o] = '\a';
			o++;
		} else if (c == 'b') {
			str[o] = '\b';
			o++;
		} else if (c == 'f') {
			str[o] = '\f';
			o++;
		} else if (c == 'v') {
			str[o] = '\v';
			o++;
		} else if (c == 'e') {
			str[o] = '\e';
			o++;
		} else if ((c >= '0') && (c <= '7')) {
			v = 0;
			while ((t < len)
			&& (str[t] >= '0') && (str[t] <= '7')) {
				v = v * 8 + (str[t] - '0');
				t++;
			}
			
			str[o] = v;
			o++;
		} else if (c == 'x') {
			t++;
			v = 0;
			while (t < len) {
				if ((str[t] >= '0') && (str[t] <= '9'))
					v = v * 16 + (str[t] - '0');
				else if ((str[t] >= 'a') && (str[t] <= 'f'))
					v = v * 16 + (str[t] - 'a');
				else if ((str[t] >= 'A') && (str[t] <= 'F'))
					v = v * 16 + (str[t] - 'A');
				else
					break;
				t++;
			}
			
			str[o] = v;
			o++;
		} else {
			str[o] = c;
			o++;
		}
	}
	
	if ((o >= 2)
	&&  (((str[0] == '\'') && (str[o - 1] == '\''))
	 ||  ((str[0] == '\"') && (str[o - 1] == '\"')))) {
		o -= 2;
		memmove(str, str + 1, o);
	}
	
	str[o] = 0;
	return(o);
}

// **************************************************************************
// 
// **************************************************************************
// Note: we need that (*str)[*len] is writable
int chop(char **str, size_t *len, char sep, char **sub_str, size_t *sub_len)
{
	char *s;
	size_t l, t, o;
	int quoted;
	
	// init
	s = *str;
	l = *len;
	
	// find length
	quoted = 0;
	
	for (t = 0; t < l; t++) {
		if (s[t] == '\\')
			continue;

		if (s[t] == '\'') {
			if (quoted == 0)
				quoted = 1;
			else if (quoted == 1)
				quoted = 0;
		} else if (s[t] == '"') {
			if (quoted == 0)
				quoted = 2;
			else if (quoted == 2)
				quoted = 0;
		} else if ((!quoted) && (s[t] == sep)) {
			s[t] = ' ';
			t++;
			break;
		}
	}
	
	// next str and len
	*str = s + t;
	*len = l - t;
	
	// remove trailing blanks
	while ((t > 0)
	&& ((s[t - 1] == ' ') || (s[t - 1] == '\t')
	 || (s[t - 1] == '\r') || (s[t - 1] == '\n')))
		t--;
	
	// remove leading blanks
	o = 0;
	while ((o < t)
	&& ((s[o] == ' ') || (s[o] == '\t')
	 || (s[o] == '\r') || (s[o] == '\n')))
		o++;
	
	// export
	s[t] = 0;

	if (sub_str)
		*sub_str = s + o;
	if (sub_len)
		*sub_len = t - o;
	
	return(t > o);
}

// **************************************************************************
// 
// **************************************************************************
void config_init(config *cfg)
{
	cfg->len = 0;
	cfg->data = NULL;
	cfg->n = 0;
	cfg->e = NULL;
}

void config_clear(config *cfg)
{
	if (cfg->data)
		xfree(cfg->data);
	if (cfg->e)
		xfree(cfg->e);
}

// **************************************************************************
// 
// **************************************************************************
static int config_parse_data(config *cfg, char sep, char eq)
{
	char *str, *es, *id, *value;
	size_t len, t, el, il, vl;
	int nmax, n;
	
	// count max entries
	str = cfg->data;
	len = cfg->len;
	nmax = 1;
	for (t = 0; t < len; t++) {
		if (str[t] == sep)
			nmax++;
	}
	
	cfg->e = xrealloc(cfg->e, nmax * sizeof(config_entry));
	
	// gather entries
	n = 0;
	while (len > 0) {
		chop(&str, &len, sep, &es, &el);
		
		if (!el)
			continue;
		
		chop(&es, &el, eq, &id, &il);
		chop(&es, &el, 0, &value, &vl);
		
		il = unescape(id, il);
		vl = unescape(value, vl);
		
		cfg->e[n].id = id;
		cfg->e[n].id_len = il;
		cfg->e[n].value = value;
		cfg->e[n].value_len = vl;
		
		//printf("%d: '%s' = '%s'\n", n, id, value);
		n++;
	}
	
	cfg->n = n;
	
	return(0);
}


// **************************************************************************
// 
// **************************************************************************
int config_parse_str(config *cfg, char *str, char sep, char eq)
{
	size_t len;
	
	// copy data
	len = strlen(str);

	cfg->data = xrealloc(cfg->data, len + 1);
	memcpy(cfg->data, str, len + 1);
	cfg->data[len] = 0;
	
	cfg->len = len;
	
	return(config_parse_data(cfg, sep, eq));
}

// **************************************************************************
// 
// **************************************************************************
int config_parse(config *cfg, char *cfg_path)
{
	int fd;
	size_t size;
	
	if (!strcmp(cfg_path, "-")) {
		fd = 0;
	} else {
		fd = open(cfg_path, O_RDONLY);
		
		if (fd < 0) {
			DEBUG("%s: %s\n", cfg_path, strerror(errno));
			return(-1);
		}
	}
	
	size = b_filesize(fd);
	
	cfg->data = xrealloc(cfg->data, size + 1);
	
	if (b_read(fd, cfg->data, size)) {
		DEBUG("%s: %s\n", cfg_path, strerror(errno));
		return(-1);
	}
	
	if (fd != 0) {
		close(fd);
	}
	
	cfg->data[size] = 0;

	cfg->len = size;
	
	return(config_parse_data(cfg, '\n', '='));
}

// **************************************************************************
// 
// **************************************************************************
/*
int config_save(config *cfg, char *cfg_path)
{
	
}
*/

// **************************************************************************
// 
// **************************************************************************
char *config_get(config *cfg, char *id, char *dflt)
{
	int t;
	
	for (t = 0; t < cfg->n; t++) {
		if (!strcmp(cfg->e[t].id, id))
			return(cfg->e[t].value);
	}
	
	return(dflt);
}

// **************************************************************************
// 
// **************************************************************************
int config_get_int(config *cfg, char *id, int dflt)
{
	char *s;
	int r;
	
	s = config_get(cfg, id, NULL);
	
	if (!s)
		return(dflt);
	
	r = strtol(s, NULL, 0);
	
	return(r);
}

// **************************************************************************
// 
// **************************************************************************
int config_get_bool(config *cfg, char *id, int dflt)
{
	char *s;
	
	s = config_get(cfg, id, NULL);
	
	if (!s)
		return(dflt);
	
	if ((s[0] == 't') || (s[0] == 'T') || (s[0] == '1'))
		return(1);
	
	if ((s[0] == 'f') || (s[0] == 'F') || (s[0] == '0'))
		return(0);
	
	return(dflt);
}

// **************************************************************************
// 
// **************************************************************************
int config_get_weekday(config *cfg, char *id, int dflt)
{
	char *s;
	
	s = config_get(cfg, id, NULL);
	
	if (!s)
		return(dflt);

	if ((s[0] >= '0') && (s[0] <= '6'))
		return(s[0] - '0');
	
	if (((s[0] == 'S') || (s[0] == 's'))
	&&  ((s[1] == 'U') || (s[1] == 'u')))
		return(0);
	if ((s[0] == 'M') || (s[0] == 'm'))
		return(1);
	if (((s[0] == 'T') || (s[0] == 't'))
	&&  ((s[1] == 'U') || (s[1] == 'u')))
		return(2);
	if ((s[0] == 'W') || (s[0] == 'w'))
		return(3);
	if (((s[0] == 'T') || (s[0] == 't'))
	&&  ((s[1] == 'H') || (s[1] == 'h')))
		return(4);
	if ((s[0] == 'F') || (s[0] == 'f'))
		return(5);
	if (((s[0] == 'S') || (s[0] == 's'))
	&&  ((s[1] == 'A') || (s[1] == 'a')))
		return(6);
	
	return(dflt);
}

// **************************************************************************
// 
// **************************************************************************
int config_get_choice(config *cfg, char *id, int dflt, int p, ...)
{
	va_list ap;
	char *s, *e;
	int t;
	
	s = config_get(cfg, id, NULL);
	
	if (!s)
		return(dflt);
	
	va_start(ap, p);
	
	for (t = 0; t < p; t++) {
		e = va_arg(ap, char *);
		
		if (!strcmp(s, e)) {
			return(t);
		}
	}
	
	va_end(ap);
	
	return(dflt);
}

// **************************************************************************
// 
// **************************************************************************
/*
void config_set(config *cfg, char *id, char *value)
{
	
}
*/

// **************************************************************************
// 
// **************************************************************************

