/*
    This file is part of "now", a simple calendar tool
    Copyright (C) 2012-2019  Laurent Poirrier

    "now" is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <string.h>
#include <stdarg.h>
#include "terminal.h"
#include "screen.h"
#include "keys.h"
#include "theme.h"
#include "memory.h"
#include "utf8.h"
#include "timeparse.h"
#include "textedit.h"
#include "dialog.h"

#define UIE_START	0
#define UIE_STOP	1
#define UIE_SUMMARY	2
#define UIE_LOCATION	3
#define UIE_DESCRIPTION	4
#define UIE_ELEMENTS	5

typedef struct uie_t {
	term *tty;
	key_list *kl;
	theme *th;
	screen *scr, *bkg;
	event *ev;
	
	textedit start, stop;
	textedit summary, location, description;
	
	int resize, active;
} uie;


// **************************************************************************
// 
// **************************************************************************
static char *xprintdiff(time_t st1, time_t st0)
{
	time_t m1, m0;
	time_t delta;
	
	m1 = dt_daymid(st1);
	m0 = dt_daymid(st0);
	
	if (m1 < m0) {
		delta = ((m0 + 12 * 3600) - m1) / (24 * 3600);
		
		if (delta > 3 * 365)
			return(xprintf("%d years earlier", delta / 365));
		if (delta > 2 * 30)
			return(xprintf("%d months earlier", delta / 30));
		if (delta > 3 * 7)
			return(xprintf("%d weeks earlier", delta / 7));
		if (delta > 1)
			return(xprintf("%d days earlier", delta));
		return(xprintf("%d day earlier", delta));
	} else {
		delta = ((m1 + 12 * 3600) - m0) / (24 * 3600);
		
		if (delta > 3 * 365)
			return(xprintf("%d years later", delta / 365));
		if (delta > 2 * 30)
			return(xprintf("%d months later", delta / 30));
		if (delta > 3 * 7)
			return(xprintf("%d weeks later", delta / 7));
		if (delta > 1)
			return(xprintf("%d days later", delta));
		return(xprintf("%d day later", delta));
	}
}

// **************************************************************************
// 
// **************************************************************************
int ui_event(term *tty, key_list *kl, theme *th, screen *scr, screen *bkg,
	event *ev)
{
	char buf[CANON_TIME_SIZE];
	char *warn, *diff;
	time_t t_start, t_stop;
	term_keypress k;
	screen_flags fl;
	uie ui;
	int b;
	int w, w2, w2p, h, h2;
	int yesno;

	ui.tty = tty;
	ui.kl = kl;
	ui.th = th;
	ui.scr = scr;
	ui.bkg = bkg;
	ui.ev = ev;
	
	ted_init(&ui.start, ui.kl, ui.scr, TED_MORELESS | TED_AT_END);
	ted_init(&ui.stop, ui.kl, ui.scr, TED_MORELESS | TED_AT_END);
	ted_init(&ui.summary, ui.kl, ui.scr, TED_AT_END);
	ted_init(&ui.location, ui.kl, ui.scr, TED_AT_END);
	ted_init(&ui.description, ui.kl, ui.scr,
		TED_MULTILINE | TED_AT_END);
	
	t_start = ev->start;
	t_stop = ev->stop;
	if (t_stop < t_start)
		t_stop = t_start;
	canon_time(buf, t_start);
	ted_set(&ui.start, buf);
	canon_duration(buf, t_start, t_stop);
	ted_set(&ui.stop, buf);
	
	ted_set(&ui.summary, ev->data[EVENT_SUMMARY]);
	ted_set(&ui.location, ev->data[EVENT_LOCATION]);
	ted_set(&ui.description, ev->data[EVENT_DESCRIPTION]);

	ui.resize = 1;
	ui.active = UIE_START;
	b = TED_CONTINUE;
	
	while (b == TED_CONTINUE) {
		if (term_signaled(ui.tty))
			ui.resize = 1;
	
		if (ui.resize) {
			if (term_getsize(ui.tty))
				exit(1);
			
			w = ui.tty->width;
			h = ui.tty->height;
			screen_resize(ui.scr, w, h);
			screen_resize(ui.bkg, w, h);

			fl = ui.th->comp[T_EV_BKG];
			
			w2 = (w - 6) / 2;
			w2p = (w & 1) ? (w2 + 1) : w2;
			h2 = (h <= 13) ? 1 : (h - 13);
			
			screen_set(ui.scr, fl, SCREEN_BLANK);
			screen_at(ui.scr, 2, 1, 20, fl, "Time:");
			screen_at(ui.scr, 2 + w2 + 2, 1, 20, fl, "End:");
			screen_at(ui.scr, 2, 4, 40, fl, "Summary:");
			screen_at(ui.scr, 2, 7, 40, fl, "Location:");
			screen_at(ui.scr, 2, 10, 40, fl, "Description:");
			
			ted_position(&ui.start, 2, 2, w2, 1);
			ted_position(&ui.stop, 2 + w2 + 2, 2, w2p, 1);
			ted_position(&ui.summary, 2, 5, w - 4, 1);
			ted_position(&ui.location, 2, 8, w - 4, 1);
			ted_position(&ui.description, 2, 11, w - 4, h2);
		}
		
		ui.resize = 0;

		fl = ui.th->comp[T_EV_EDITABLE];
		ted_paint(&ui.start, fl, ui.active == UIE_START);
		ted_paint(&ui.stop, fl, ui.active == UIE_STOP);
		ted_paint(&ui.summary, fl, ui.active == UIE_SUMMARY);
		ted_paint(&ui.location, fl, ui.active == UIE_LOCATION);
		ted_paint(&ui.description, fl, ui.active == UIE_DESCRIPTION);

		screen_tup(ui.scr, ui.tty, ui.bkg);

		ted_cursor_paint(&ui.start, ui.tty, ui.active == UIE_START);
		ted_cursor_paint(&ui.stop, ui.tty, ui.active == UIE_STOP);
		ted_cursor_paint(&ui.summary, ui.tty, ui.active == UIE_SUMMARY);
		ted_cursor_paint(&ui.location, ui.tty,
			ui.active == UIE_LOCATION);
		ted_cursor_paint(&ui.description, ui.tty,
			ui.active == UIE_DESCRIPTION);

		term_getkey(ui.tty, &k);
	
		if (ui.active == UIE_START) {
			b = ted_events(&ui.start, &k);
		} else if (ui.active == UIE_STOP) {
			b = ted_events(&ui.stop, &k);
		} else if (ui.active == UIE_SUMMARY) {
			b = ted_events(&ui.summary, &k);
		} else if (ui.active == UIE_LOCATION) {
			b = ted_events(&ui.location, &k);
		} else if (ui.active == UIE_DESCRIPTION) {
			b = ted_events(&ui.description, &k);
		}
		
		if (b == TED_NEXT) {
			ui.active++;
			if (ui.active >= UIE_ELEMENTS)
				ui.active = 0;
			
			b = TED_CONTINUE;
		} else if (b == TED_PREV) {
			ui.active--;
			if (ui.active < 0)
				ui.active = UIE_ELEMENTS - 1;
			
			b = TED_CONTINUE;
		} else if (((b == TED_MORE) || (b == TED_LESS))
		&& (ui.active == UIE_START)) {
			t_start = parse_time(ev->start, ted_get(&ui.start));

			if (t_start != (time_t)-1) {
				if (b == TED_LESS)
					t_start -= 3600;
				else
					t_start += 3600;
				
				canon_time(buf, t_start);
				ted_set(&ui.start, buf);
				ted_set_modified(&ui.start, 1);
			}
			
			b = TED_CONTINUE;
		} else if (((b == TED_MORE) || (b == TED_LESS))
		&& (ui.active == UIE_STOP)) {
			t_start = parse_time(ev->start, ted_get(&ui.start));
			t_stop = parse_duration(t_start, ted_get(&ui.stop));
			
			if ((t_start != (time_t)-1) && (t_stop != (time_t)-1)) {
				if (b == TED_LESS)
					t_stop -= 1800;
				else
					t_stop += 1800;
				
				if (t_stop < t_start)
					t_stop = t_start;
				
				canon_duration(buf, t_start, t_stop);
				ted_set(&ui.stop, buf);
				ted_set_modified(&ui.start, 1);
			}
			
			b = TED_CONTINUE;
		} else if ((b == TED_CONFIRM)
		&& (ui.active == UIE_DESCRIPTION)) {
			ui.active = UIE_START;
			
			b = TED_CONTINUE;
		} else if (b == TED_CONFIRM) {
			t_start = parse_time(ev->start, ted_get(&ui.start));
			t_stop = parse_duration(t_start, ted_get(&ui.stop));
		
			if (t_start == (time_t)-1) {
				ui.active = UIE_START;
				b = TED_CONTINUE;
			} else if ((t_stop == (time_t)-1)
			|| (t_stop < t_start)) {
				ui.active = UIE_STOP;
				b = TED_CONTINUE;
			} else if (ted_get(&ui.summary)[0] == 0) {
				ui.active = UIE_SUMMARY;
				b = TED_CONTINUE;
			} else if (dt_daymid(t_start)
			!= dt_daymid(ev->start)) {
				diff = xprintdiff(t_start, ev->start);
				warn = xprintf("You are moving the event %s:",
					diff);

				yesno = ui_selectn(ui.tty, ui.kl, ui.th,
					ui.scr, ui.bkg,
					warn, 2,
					"Yes, this is intended.",
					"No, continue editing.");
				
				xfree(warn);
				xfree(diff);
				
				ui.resize = 1;
				
				if (yesno != 0)
					b = TED_CONTINUE;
			}
		} else if ((b == TED_CANCEL)
		&& ((ted_modified(&ui.start)) || (ted_modified(&ui.stop))
		||  (ted_modified(&ui.summary)) || (ted_modified(&ui.location))
		||  (ted_modified(&ui.description)))) {
			yesno = ui_selectn(ui.tty, ui.kl, ui.th,
				ui.scr, ui.bkg,
				"You modified the event:", 2,
				"Discard changes.",
				"Continue editing.");
			ui.resize = 1;
			
			if (yesno != 0)
				b = TED_CONTINUE;
		}
	}
	
	if (b == TED_CONFIRM) {
		ev->start = t_start;
		ev->stop = t_stop;
		event_set(ev, EVENT_SUMMARY, ted_get(&ui.summary));
		event_set(ev, EVENT_LOCATION, ted_get(&ui.location));
		event_set(ev, EVENT_DESCRIPTION, ted_get(&ui.description));
	}
	
	ted_clear(&ui.start);
	ted_clear(&ui.stop);
	ted_clear(&ui.summary);
	ted_clear(&ui.location);
	ted_clear(&ui.description);
	return(b != TED_CONFIRM);
}

// **************************************************************************
// 
// **************************************************************************
int ui_selectn(term *tty, key_list *kl, theme *th, screen *scr, screen *bkg,
	char *title, int n, ...)
{
	char *tab[n];
	va_list ap;
	int t;
	
	va_start(ap, n);
	for (t = 0; t < n; t++) {
		tab[t] = va_arg(ap, char *);
	}
	va_end(ap);
	
	return(ui_select(tty, kl, th, scr, bkg, title, n, tab));
}

int ui_select(term *tty, key_list *kl, theme *th, screen *scr, screen *bkg,
	char *title, int n, char **opts)
{
	term_keypress k;
	screen_flags fl;
	int resize, active, offset, b;
	int w, h, ey, eh, t, l;
	
	offset = 0;
	active = 0;
	resize = 1;
	w = h = 0;
	ey = eh = 0;
	active = UIE_START;
	b = TED_CONTINUE;
	
	while (b == TED_CONTINUE) {
		if (term_signaled(tty))
			resize = 1;
	
		if (resize) {
			if (term_getsize(tty))
				exit(1);
			
			w = tty->width;
			h = tty->height;
			screen_resize(scr, w, h);
			screen_resize(bkg, w, h);
			
			screen_set(scr, th->comp[T_SR_BKG], SCREEN_BLANK);
			screen_at(scr, 2, 1, w - 4, th->comp[T_SR_TITLE],
				title);
			
			ey = 4;
			eh = h - 5;
		}
		
		resize = 0;

		if (active > n - 1)
			active = n - 1;
		if (active < 0)
			active = 0;
		
		if (active > offset + eh - 1)
			offset = active - eh + 1;
		if (active < offset)
			offset = active;

		for (t = 0; t < n; t++) {
			fl = (t == active)
				? th->comp[T_SR_SEL] : th->comp[T_SR_BKG];
			
			l = ey + t - offset;
			
			if ((l >= ey) && (l < ey + eh))
				screen_at(scr, 2, l, w - 4, fl, opts[t]);
		}
		
		screen_at(scr, 2, ey - 1, w - 4, th->comp[T_SR_BKG],
			(offset > 0) ? "..." : "   ");
		screen_at(scr, 2, ey + eh, w - 4, th->comp[T_SR_BKG],
			(n - offset > eh) ? "..." : "   ");
		
		screen_tup(scr, tty, bkg);
		term_goto(tty, w - 1, h - 1);

		term_getkey(tty, &k);
	
		if (key_action(kl, &k, A_SEL_CANCEL)) {
			b = TED_CANCEL;
		} else if (key_action(kl, &k, A_SEL_CONFIRM)) {
			b = TED_CONFIRM;
		} else if (key_action(kl, &k, A_SEL_PREV)) {
			active--;
		} else if (key_action(kl, &k, A_SEL_NEXT)) {
			active++;
		} else if (key_action(kl, &k, A_SEL_M10)) {
			active -= 10;
		} else if (key_action(kl, &k, A_SEL_P10)) {
			active += 10;
		} else if (key_action(kl, &k, A_SEL_FIRST)) {
			active = 0;
		} else if (key_action(kl, &k, A_SEL_LAST)) {
			active = 1 << 24;
		}
	}
	
	if (b == TED_CONFIRM) {
		return(active);
	}
	
	return(-1);
}


