/*
    This file is part of "now", a simple calendar tool
    Copyright (C) 2012-2019  Laurent Poirrier

    "now" is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <string.h>
#include "memory.h"
#include "event.h"

// ***************************************************************************
// 
// ***************************************************************************
event *event_new()
{
	event *e;
	int line;

	e = xmalloc(sizeof(event));
	
	e->start = e->stop = 0;
	
	for (line = 0; line < EVENT_LINES; line++) {
		e->data[line] = NULL;
	}
	return(e);
}

// ***********************************************************
void event_set(event *e, int line, char *content)
{
	size_t l;
	
	if ((line < 0) || (line >= EVENT_LINES))
		return;
	
	if (e->data[line])
		xfree(e->data[line]);
	
	if (content) {
		l = strlen(content);
	
		e->data[line] = xmalloc(l + 1);
	
		memcpy(e->data[line], content, l + 1);
	} else {
		e->data[line] = NULL;
	}
}

// ***********************************************************
void event_copy(event *dst, event *src)
{
	int line;
	
	dst->start = src->start;
	dst->stop = src->stop;
	
	for (line = 0; line < EVENT_LINES; line++) {
		event_set(dst, line, src->data[line]);
	}
}

// ***********************************************************
void event_destroy(event *e)
{
	int line;
	
	for (line = 0; line < EVENT_LINES; line++) {
		if (e->data[line])
			xfree(e->data[line]);
	}
	
	xfree(e);
}


// ***************************************************************************
// 
// ***************************************************************************
void evlist_init(event_list *l)
{
	l->size = 0;
	l->head = l->tail = NULL;
}

// ***********************************************************
void evlist_clear(event_list *l)
{
	event_handle *h, *h2;
	
	h = l->head;
	while (h) {
		h2 = h->next;
		xfree(h);
		h = h2;
	}
	
	l->head = l->tail = NULL;
	l->size = 0;
}

// ***********************************************************
void evlist_destroy_all(event_list *l)
{
	event_handle *h, *h2;
	
	h = l->head;
	while (h) {
		h2 = h->next;
		event_destroy(h->ev);
		xfree(h);
		h = h2;
	}
	
	l->head = l->tail = NULL;
	l->size = 0;
}

// ***********************************************************
event_handle *evlist_attach(event_list *l, event *e)
{
	event_handle *h;
	
	h = xmalloc(sizeof(event_handle));
	
	h->ev = e;
	h->orig = h;

	if (!l->tail) {
		l->head = l->tail = h;
		h->prev = h->next = NULL;
		l->size = 1;
		return(h);
	}
	
	h->prev = l->tail;
	h->next = NULL;
	
	l->tail->next = h;
	
	l->tail = h;
	l->size++;
	return(h);
}

// ***********************************************************
void evlist_detach(event_list *l, event_handle *h)
{
	if (h->prev) {
		h->prev->next = h->next;
	} else {
		l->head = h->next;
	}
	
	if (h->next) {
		h->next->prev = h->prev;
	} else {
		l->tail = h->prev;
	}
	
	l->size--;
	xfree(h);
}

// ***********************************************************
void evlist_select(event_list *dst, event_list *src, time_t begin, time_t end)
{
	event_handle *h, *n;
	
	h = src->head;
	
	while (h) {
		if ((h->ev->start < end) && (h->ev->stop > begin)) {
			n = evlist_attach(dst, h->ev);
			n->orig = h->orig;
		}
		h = h->next;
	}
}

// ***********************************************************
void evlist_slowsort(event_list *l)
{
	event_handle *top, *h, *n;
	
	top = l->tail;
	while (top != l->head) {
		h = l->head;
		while (h != top) {
			n = h->next;
			
			if (h->ev->start > n->ev->start) {
				if (h->prev)
					h->prev->next = n;
				else
					l->head = n;
				
				if (n->next)
					n->next->prev = h;
				else
					l->tail = h;
				
				h->next = n->next;
				n->prev = h->prev;
				n->next = h;
				h->prev = n;
				
				if (top == n)
					top = h;
			} else {
				h = n;
			}
		}
		
		top = top->prev;
	}
}


// ***************************************************************************
// 
// ***************************************************************************


