/*
    This file is part of "now", a simple calendar tool
    Copyright (C) 2012-2019  Laurent Poirrier

    "now" is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "bio.h"
#include "memory.h"
#include "event.h"
#include "ical.h"

#define DEBUG(...)	fprintf(stderr, __VA_ARGS__)

// ***************************************************************************
// 
// ***************************************************************************
static char *shrink(char *s, char **sep_r)
{
	size_t i, o;
	char *sep;
	
	if (!s)
		return(NULL);
	
	sep = NULL;
	i = o = 0;
	
	while (1) {
		if (s[i] == 0) {
			s[o] = 0;
			if (sep_r)
				*sep_r = sep;
			return(&s[i]);
		} else if (s[i] == '\r') {
			i++;
		} else if (s[i] == '\n') {
			if (s[i + 1] == ' ') {
				i += 2;
			} else {
				s[o] = 0;
				if (sep_r)
					*sep_r = sep;
				return(&s[i + 1]);
			}
		} else if (s[i] == '\\') {
			if ((s[i + 1] == 'n') || (s[i + 1] == 'N')) {
				s[o] = '\n';
				i += 2;
				o++;
			} else if ((s[i + 1] == 't') || (s[i + 1] == 'T')) {
				s[o] = '\t';
				i += 2;
				o++;
			} else if (s[i + 1] == ',') {
				s[o] = ',';
				i += 2;
				o++;
			} else if (s[i + 1] == ';') {
				s[o] = ';';
				i += 2;
				o++;
			} else {
				s[o] = '\\';
				i++;
				o++;
			}
		} else {
			if ((!sep) && (s[i] == ':'))
				sep = &s[i + 1];
			s[o] = s[i];
			i++;
			o++;
		}
	}
}

// ***************************************************************************
// 
// ***************************************************************************
static void fprintical(FILE *f, char *h, char *s)
{
	char buff[96];
	size_t o;
	
	o = strlen(h);
	if (o >= 70) {
		DEBUG("BUG: header %s too long\n", h);
		exit(1);
	}
	
	memcpy(buff, h, o);
	buff[o++] = ':';
	
	while (*s) {
		if (o + 2 > 75) {
			buff[o++] = '\r';
			buff[o++] = '\n';
			fwrite(buff, o, 1, f);
			buff[0] = ' ';
			o = 1;
		}
		
		if (*s == '\n') {
			buff[o++] = '\\';
			buff[o++] = 'n';
		} else if (*s == '\r') {
			buff[o++] = '\\';
			buff[o++] = 'r';
		} else if (*s == '\t') {
			buff[o++] = '\\';
			buff[o++] = 't';
		} else if (*s == ',') {
			buff[o++] = '\\';
			buff[o++] = ',';
		} else if (*s == ';') {
			buff[o++] = '\\';
			buff[o++] = ';';
		} else {
			buff[o++] = *s;
		}

		s++;
	}

	buff[o++] = '\r';
	buff[o++] = '\n';
	fwrite(buff, o, 1, f);
}

// ***************************************************************************
// 
// ***************************************************************************
time_t ical_time(char *s)
{
	dt_time dt;
	int buf[16] = { 1, 9, 7, 0, 0, 1, 0, 1, -1, 0, 0, 0, 0, 0, 0, -1 };
	size_t l;
	time_t r;
	
	if (!s)
		return(0);
	
	while ((*s == ' ') || (*s == '\n') || (*s == '\r') || (*s == '\t'))
		s++;
	
	l = 0;
	while ((l < 16) && (s[l] != 0)) {
		buf[l] = s[l] - '0';
		l++;
	}
	
	dt.isdst = -1;
	dt.year = buf[0] * 1000 + buf[1] * 100 + buf[2] * 10 + buf[3];
	dt.mon = buf[4] * 10 + buf[5];
	dt.mday = buf[6] * 10 + buf[7];
	dt.hour = buf[9] * 10 + buf[10];
	dt.min = buf[11] * 10 + buf[12];
	dt.sec = buf[13] * 10 + buf[14];
	
	r = dt_mktime(&dt);
	
	return(r);
}

// ***************************************************************************
// 
// ***************************************************************************
static event *ical_event(char *start, char *stop,
	char *sum, char *loc, char *desc)
{
	
	event *e;
	
	e = event_new();
	
	e->start = ical_time(start);
	e->stop = ical_time(stop);
	event_set(e, EVENT_SUMMARY, sum);
	event_set(e, EVENT_LOCATION, loc);
	event_set(e, EVENT_DESCRIPTION, desc);
	
	/*{
		static int cnt = 0;

		dt_time t[2];
		int c;
		
		
		printf("----- %d\n", cnt++);

		dt_localtime(e->start, &t[0]);
		dt_localtime(e->stop, &t[1]);
		
		for (c = 0; c <= 1; c++) {
			printf("time%d: %d-%d-%d %d:%d:%d\n", c,
				t[c].year, t[c].mon, t[c].mday,
				t[c].hour, t[c].min, t[c].sec);
		}

		printf("Summary: %s\n", sum);
	
		if (loc)
			printf("Location: %s\n", loc);
	
		if (desc)
			printf("Description:\n%s\n", desc);
	}*/
	return(e);
}

// ***************************************************************************
// 
// ***************************************************************************
static int evlist_ical_parse(event_list *l, size_t size, char *data)
{
	int mode = 0;
	char *lb, *le, *sep;
	
	char *start, *stop;
	char *sum, *loc, *desc;
	event *e;
	
	lb = data;
	while (*lb != 0) {
		le = lb;
		sep = NULL;

		le = shrink(lb, &sep);
		
		if (mode == 0) {
			if (!(strncmp(lb, "BEGIN:VCALENDAR", 15))) {
				mode = 1;
			}
		} else if (mode == 1) {
			if (!(strncmp(lb, "END:VCALENDAR", 13))) {
				mode = 0;
			} else if (!strncmp(lb, "BEGIN:VEVENT", 12)) {
				mode = 2;
				start = stop = NULL;
				sum = loc = desc = NULL;
			}
		} else if (mode == 2) {
			if (!(strncmp(lb, "END:VEVENT", 10))) {
				e = ical_event(start, stop, sum, loc, desc);
				if (e) {
					evlist_attach(l, e);
				}
				mode = 1;
			} else if ((sep) && (!strncmp(lb, "DTSTART", 7))) {
				start = sep;
			} else if (!strncmp(lb, "DTEND", 5)) {
				stop = sep;
			} else if (!strncmp(lb, "SUMMARY", 7)) {
				sum = sep;
			} else if (!strncmp(lb, "LOCATION", 8)) {
				loc = sep;
			} else if (!strncmp(lb, "DESCRIPTION", 11)) {
				desc = sep;
			}
		}
		
		lb = le;
	}
	
	if (lb != data + size)
		return(-1);

	return(0);
}

// ***************************************************************************
// 
// ***************************************************************************
int evlist_ical_load(event_list *l, char *path)
{
	size_t size;
	char *data;
	int fd;
	
	fd = open(path, O_RDONLY);
	if (fd < 0) {
		DEBUG("%s: %s\n", path, strerror(errno));
		return(-1);
	}
	
	size = b_filesize(fd);
	if (size == 0) {
		DEBUG("%s: no data\n", path);
		close(fd);
		return(-1);
	}
	
	data = xmalloc(size + 1);
	
	if (b_read(fd, data, size)) {
		DEBUG("%s: read error\n", path);
		xfree(data);
		close(fd);
		return(-1);
	}
	
	data[size] = 0;
	
	if (evlist_ical_parse(l, size, data)) {
		DEBUG("%s: parse error\n", path);
		xfree(data);
		close(fd);
		return(-1);
	}
	
	xfree(data);
	close(fd);
	return(0);
}

// ***************************************************************************
// 
// ***************************************************************************
static int event_ical_write(event *e, FILE *f)
{
	dt_time t[2];
	char *tn[2] = { "DTSTART", "DTEND" };
	char *dn[3] = { "SUMMARY", "LOCATION", "DESCRIPTION" };
	int c;
	
	fprintf(f, "BEGIN:VEVENT\r\n");
	
	dt_localtime(e->start, &(t[0]));
	dt_localtime(e->stop, &(t[1]));
	
	for (c = 0; c < 2; c++) {
		fprintf(f, "%s:%04d%02d%02dT%02d%02d%02d\r\n",
			tn[c], t[c].year, t[c].mon, t[c].mday,
			t[c].hour, t[c].min, t[c].sec);
	}
	
	for (c = 0; c < 3; c++) {
		if (e->data[c]) {
			fprintical(f, dn[c], e->data[c]);
		}
	}
	
	fprintf(f, "END:VEVENT\r\n");
	return(ferror(f));
}

// ************************************************************
static int evlist_ical_write(event_list *l, FILE *f)
{
	event_handle *h;
	
	h = l->head;
	
	fprintf(f, "BEGIN:VCALENDAR\r\n");
	
	while (h) {
		if (event_ical_write(h->ev, f))
			return(-1);
		
		h = h->next;
	}

	fprintf(f, "END:VCALENDAR\r\n");
	
	return(ferror(f));
}

// ***************************************************************************
// 
// ***************************************************************************
int evlist_ical_save(event_list *l, char *path)
{
	FILE *f;
	int r;
	
	f = fopen(path, "w");
	if (!f) {
		DEBUG("%s: %s\n", path, strerror(errno));
		return(-1);
	}
	
	r = evlist_ical_write(l, f);
	
	fclose(f);
	return(r);
}


