/*
    This file is part of "now", a simple calendar tool
    Copyright (C) 2012-2019  Laurent Poirrier

    "now" is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include "config.h"
#include "keys.h"

// **************************************************************************
// 
// **************************************************************************
#define N_SHORTCUTS	((int)(sizeof(shortcuts) / sizeof(shortcuts[0])))
static struct {
	char *name;
	char *pattern;
} shortcuts[] = {
	{ "escape", "\033" },
	{ "alt+escape", "\033\033" },
	
	{ "up", "\033[A" },
	{ "down", "\033[B" },
	{ "left", "\033[D" },
	{ "right", "\033[C" },
	{ "+up", "\033[1;#A" },
	{ "+down", "\033[1;#B" },
	{ "+left", "\033[1;#D" },
	{ "+right", "\033[1;#C" },
	
	{ "+insert", "\033[2;#~" },
	{ "+delete", "\033[3;#~" },
	{ "+page-up", "\033[5;#~" },
	{ "+page-down", "\033[6;#~" },
	
	{ "home", "\033[H" },
	{ "+home", "\033[1;#H" },
	{ "+home", "\033O#H" },
	{ "+home", "\033[1;#~" },
	{ "+home", "\033[7;#~" },
	{ "end", "\033[F" },
	{ "+end", "\033[1;#F" },
	{ "+end", "\033O#F" },
	{ "+end", "\033[4;#~" },
	{ "+end", "\033[8;#~" },
	
	{ "backspace", "\177" },
	{ "alt+backspace", "\033\177" },
	{ "enter", "\n" },
	{ "shift+enter", "\033OM" },
	{ "alt+enter", "\033\n" },
	{ "alt+shift+enter", "\033\033OM" },
	{ "space", " " },
	{ "control+space", "(\000)" },
	{ "tab", "\t" },
	{ "shift+tab", "\033[Z" },

	{ "+f1", "\033O#P"},
	{ "+f2", "\033O#Q"},
	{ "+f3", "\033O#R"},
	{ "+f4", "\033O#S"},
	{ "+f5", "\033[15;#~"},
	{ "+f6", "\033[17;#~"},
	{ "+f7", "\033[18;#~"},
	{ "+f8", "\033[19;#~"},
	{ "+f9", "\033[20;#~"},
	{ "+f10", "\033[21;#~"},
	{ "+f11", "\033[23;#~"},
	{ "+f12", "\033[24;#~"},
};

#define MODIF_NONE	0
#define MODIF_SUPER	1
#define MODIF_SHIFT	2
#define MODIF_ALT	4
#define MODIF_CONTROL	8

// **************************************************************************
// 
// **************************************************************************
static struct {
	char *name;
	char *descr;
} actions[A_ACTIONS] = {
	{ "cal-exit", "escape, control+c" },
	{ "cal-enter", "enter" },
	{ "cal-week-prev", "up" },
	{ "cal-week-next", "down" },
	{ "cal-shift-m1", "alt+up" },
	{ "cal-shift-p1", "alt+down" },
	{ "cal-shift-m4", "page-up" },
	{ "cal-shift-p4", "page-down" },
	{ "cal-day-prev", "left" },
	{ "cal-day-next", "right" },
	{ "cal-week-start", "home" },
	{ "cal-week-end", "end" },
	{ "cal-week-start-m1", "alt+left" },
	{ "cal-week-start-p1", "alt+right" },
	
	{ "day-leave", "escape, control+c" },
	{ "day-ev-prev", "up" },
	{ "day-ev-next", "down" },
	{ "day-day-prev", "left, alt+left" },
	{ "day-day-next", "right, alt+right" },
	{ "day-ev-first", "home" },
	{ "day-ev-last", "end" },
	{ "day-shift-m1", "alt+up" },
	{ "day-shift-p1", "alt+down" },
	{ "day-shift-m4", "page-up" },
	{ "day-shift-p4", "page-down" },
	{ "day-edit", "enter" },
	{ "day-delete", "delete, backspace" },
	
	{ "sel-cancel", "escape, control+c" },
	{ "sel-confirm", "enter" },
	{ "sel-prev", "up" },
	{ "sel-next", "down" },
	{ "sel-m10", "page-up" },
	{ "sel-p10", "page-down" },
	{ "sel-first", "home" },
	{ "sel-last", "end" },
	
	{ "ed-single-exit", "escape" },
	{ "ed-single-confirm", "enter" },
	{ "ed-single-prev", "shift+tab, up" },
	{ "ed-single-next", "tab, down" },
	{ "ed-multi-exit", "" },
	{ "ed-multi-confirm", "escape, shift+enter" },
	{ "ed-multi-prev", "shift+tab" },
	{ "ed-multi-next", "" },
	{ "ed-multi-first-prev", "up" },
	{ "ed-multi-up", "up" },
	{ "ed-multi-down", "down" },
	{ "ed-multi-up-10", "page-up" },
	{ "ed-multi-down-10", "page-down" },
	{ "ed-more", "page-up" },
	{ "ed-less", "page-down" },
	{ "ed-left", "left" },
	{ "ed-right", "right" },
	{ "ed-home", "home" },
	{ "ed-end", "end" },
	{ "ed-delete", "delete" },
	{ "ed-backspace", "backspace" },
	{ "ed-newline", "enter" },
	{ "ed-tab", "tab" },
};

// **************************************************************************
// 
// **************************************************************************
void key_init(key_list *kl)
{
	memset(kl->action, 0,
		A_ACTIONS * KEY_ACTION_BINDINGS * sizeof(key_str));
}

void key_clear(key_list *kl)
{
	(void)kl;
}

// **************************************************************************
// 
// **************************************************************************
static size_t key_str_build(char *dst, int modif, char *str)
{
	char *e;
	int k;
	size_t i, o, len;
	
	if (str[0] == '(') {
		str++;
		e = memchr(str, ')', TERM_KEY_STR_LENGTH + 1);
		len = (e) ? e - str : TERM_KEY_STR_LENGTH;
	} else {
		len = strlen(str);
		
		if (len > TERM_KEY_STR_LENGTH)
			len = TERM_KEY_STR_LENGTH;
	}
	
	o = 0;
	for (i = 0; i < len; i++) {
		if (str[i] == ';') {
			if (modif != MODIF_NONE) {
				dst[o] = ';';
				o++;
			}
		} else if (str[i] == '#') {
			if (modif != MODIF_NONE) {
				k = 1;
				if (modif & MODIF_SHIFT)
					k += 1;
				if (modif & MODIF_ALT)
					k += 2;
				if (modif & MODIF_CONTROL)
					k += 4;
				if (modif & MODIF_SUPER)
					k = 1;

				dst[o] = '0' + k;
				o++;
			}
		} else {
			dst[o] = str[i];
			o++;
		}
	}
	
	return(o);
}

// **************************************************************************
// 
// **************************************************************************
static int key_binding_parse(key_str *action, int max, char *fs, size_t fl)
{
	char *ss;
	size_t sl, len, ofs;
	int b, t;
	int modif;
	char c;
	
	// parse modifiers
	ss = fs;
	sl = fl;
	modif = MODIF_NONE;
	while (1) {
		if (!strncmp(ss, "super+", 6)) {
			modif |= MODIF_SUPER;
			ss += 6;
			sl -= 6;
		} else if (!strncmp(ss, "shift+", 6)) {
			modif |= MODIF_SHIFT;
			ss += 6;
			sl -= 6;
		} else if (!strncmp(ss, "alt+", 4)) {
			modif |= MODIF_ALT;
			ss += 4;
			sl -= 4;
		} else if (!strncmp(ss, "control+", 8)) {
			modif |= MODIF_CONTROL;
			ss += 8;
			sl -= 8;
		} else {
			break;
		}
	}
	
	// match shortcuts
	b = 0;
	for (t = 0; t < N_SHORTCUTS; t++) {
		if (shortcuts[t].name[0] == '+') {
			if (strcmp(shortcuts[t].name + 1, ss))
				continue;
		} else {
			if (strcmp(shortcuts[t].name, fs))
				continue;
		}
		
		if (b >= max)
			return(b);
		
		action[b].len = key_str_build(action[b].str,
			modif, shortcuts[t].pattern);
		b++;
	}
	
	if (b > 0)
		return(b);

	// alt prefix
	ofs = 0;
	if (modif & MODIF_ALT) {
		action[b].str[0] = '\033';
		ofs++;
	}
	
	// control+letter
	if ((modif & MODIF_CONTROL) && (ss[0] != 0) && (ss[1] == 0)) {
		c = ss[0];
		if ((c >= 'A') && (c <= 'Z'))
			c = c - 'A' + 1;
		else if ((c >= 'a') && (c <= 'z'))
			c = c - 'a' + 1;
		else
			return(b);
		
		action[b].str[ofs] = c;
		action[b].len = ofs + 1;
		return(b + 1);
	}
	
	// raw string, escaped
	len = unescape(ss, sl);
	if (ofs + len > TERM_KEY_STR_LENGTH)
		return(b);
	memcpy(&action[b].str[ofs], ss, len);
	action[b].len = ofs + len;

	return(b + 1);
}

// **************************************************************************
// 
// **************************************************************************
#define KEY_PARSE_MAX	256
static void key_action_parse(key_str *action, int max, char *descr)
{
	char buff[KEY_PARSE_MAX];
	char *s, *ss;
	size_t len, l, sl;
	int b;
	
	// local copy
	len = strlen(descr);
	if (len >= KEY_PARSE_MAX)
		len = KEY_PARSE_MAX - 1;
	memcpy(buff, descr, len);
	buff[len] = 0;
	s = buff;
	l = len;
	
	// chop
	b = 0;
	while ((l > 0) && (b < max)) {
		chop(&s, &l, ',', &ss, &sl);
		
		if (sl == 0)
			continue;
		
		b += key_binding_parse(&action[b], max - b, ss, sl);
	}
}

// **************************************************************************
// 
// **************************************************************************
void key_config(key_list *kl, config *cfg)
{
	char *descr;
	int a;
	
	for (a = 0; a < A_ACTIONS; a++) {
		descr = NULL;
		
		if (cfg)
			descr = config_get(cfg, actions[a].name, NULL);
		
		if (!descr)
			descr = actions[a].descr;
		
		key_action_parse(kl->action[a], KEY_ACTION_BINDINGS,
			descr);

		// debug
		/*printf("%s: [default %s] %s: \n",
			actions[a].name, actions[a].descr, descr);
		
		for (int b = 0; b < KEY_ACTION_BINDINGS; b++) {
			if (kl->action[a][b].len == 0)
				continue;
			
			printf("    ");
			term_key_dump(kl->action[a][b].str,
				kl->action[a][b].len, 1);
		}*/
	}
}

// **************************************************************************
// 
// **************************************************************************
int key_action(key_list *kl, term_keypress *k, int a)
{
	int b;
	
	for (b = 0; b < KEY_ACTION_BINDINGS; b++) {
		if (kl->action[a][b].len == 0)
			continue;
		
		if (term_match(k,
				kl->action[a][b].str,
				kl->action[a][b].len))
			return(1);
	}
	
	return(0);
}


// **************************************************************************
// 
// **************************************************************************
int key_name(char *buff, size_t size, term_keypress *k)
{
	char str[TERM_KEY_STR_LENGTH + 1];
	size_t len, w;
	char *s;
	int t, modif, modif_last;
	
	// shortcuts
	for (t = 0; t < N_SHORTCUTS; t++) {
		s = shortcuts[t].name;
		if (s[0] == '+') {
			s++;
			modif_last = MODIF_SUPER | MODIF_SHIFT | MODIF_ALT
				| MODIF_CONTROL;
		} else {
			modif_last = MODIF_NONE;
		}
		
		for (modif = MODIF_NONE; modif <= modif_last; modif++) {
			len = key_str_build(str, modif, shortcuts[t].pattern);
			
			if (term_match(k, str, len)) {
				w = snprintf(buff, size, "%s%s%s%s%s",
					(modif & MODIF_CONTROL)
						? "control+" : "",
					(modif & MODIF_ALT) ? "alt+" : "",
					(modif & MODIF_SHIFT) ? "shift+" : "",
					(modif & MODIF_SUPER) ? "super+" : "",
					s);
				
				if (w < size)
					return(0);
				if (size >= 1)
					buff[size - 1] = 0;
				return(-1);
			}
		}
	}
	
	// alt+, control+
	len = k->n;
	
	if (len > TERM_KEY_STR_LENGTH)
		len = TERM_KEY_STR_LENGTH;
	
	memcpy(str, k->k, len);
	str[len] = 0;
	s = str;
	
	modif = 0;

	if ((len >= 2) && (s[0] == '\033')) {
		s++;
		len--;
		modif |= MODIF_ALT;
	}
	
	if ((len == 1) && ((s[0] >= 1) && (s[0] <= 26))) {
		s[0] = s[0] - 1 + 'a';
		modif |= MODIF_CONTROL;
	}
	
	w = snprintf(buff, size, "%s%s%s",
		(modif & MODIF_CONTROL) ? "control+" : "",
		(modif & MODIF_ALT) ? "alt+" : "",
		s);
	
	if (w < size)
		return(0);
	
	if (size >= 1)
		buff[size - 1] = 0;
	return(-1);
}

// ****************************************************************************
// 
// ****************************************************************************
int key_escape(char *buff, size_t size, term_keypress *k)
{
	char *s;
	size_t l, i, o;
	
	s = k->k;
	l = k->n;
	
	o = 0;
	for (i = 0; i < l; i++) {
		if (((s[i] >= 0) && (s[i] < 32)) || (s[i] == 127)) {
			if (o + 0 < size)
				buff[o + 0] = '\\';
			if (o + 1 < size)
				buff[o + 1] = '0' + ((s[i] >> 6) & 7);
			if (o + 2 < size)
				buff[o + 2] = '0' + ((s[i] >> 3) & 7);
			if (o + 3 < size)
				buff[o + 3] = '0' + (s[i] & 7);
			o += 4;
		} else {
			if (o < size)
				buff[o] = s[i];
			o++;
		}
	}
	
	if (o < size) {
		buff[o] = 0;
		return(0);
	}
	
	if (size > 0)
		buff[size - 1] = 0;
	
	return(-1);
}

