/*
    This file is part of "now", a simple calendar tool
    Copyright (C) 2012-2019  Laurent Poirrier

    "now" is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "memory.h"
#include "config.h"
#include "bio.h"
#include "dateproc.h"
#include "utf8.h"
#include "terminal.h"
#include "screen.h"
#include "event.h"
#include "ical.h"
#include "ui.h"
#include "textedit.h"
#include "main.h"

// *************************************************************************
// 
// *************************************************************************
#define DEBUG(...)	fprintf(stderr, __VA_ARGS__)
#define PRINT(...)	fprintf(stdout, __VA_ARGS__)

#ifndef NOW_VERSION
#define NOW_VERSION	"X.XX"
#endif

#ifndef NOW_YEARS
#define NOW_YEARS	"2012-"
#endif


// *************************************************************************
// 
// *************************************************************************
#define HELP_MSG	\
"Usage: now [-c file.ics]\n" \
"\n" \
"Options:\n" \
"  -f, --config         Specify configuration file\n" \
"                       (default: $HOME/.config/now.conf)\n" \
"  -c, --calendar       Specify iCal file\n" \
"                       (default: $HOME/.local/share/now/calendar.ics)\n" \
"  -h, --help           Show this help and exit\n" \
"  -v, --version        Show version and exit\n" \
"\n"

#define VERSION_MSG	\
"now v" NOW_VERSION " Copyright (C) " NOW_YEARS "  Laurent Poirrier\n" \
"\n" \
"This program comes with ABSOLUTELY NO WARRANTY.\n" \
"This is free software, and you are welcome to redistribute it\n" \
"under certain conditions; see COPYING for details.\n" \
"\n"

// *************************************************************************
// 
// *************************************************************************
/* static int test_daytime(char *cal_path)
{
	dt_time r;
	time_t t;
	
	r.year = 2000 + 12;
	r.mon = 3;
	r.mday = 23;
	r.hour = 12;
	r.min = 0;
	r.sec = 0;
	
	t = dt_nextday(dt_mktime(&r), 1);
	
	dt_localtime(t, &r);
	
	printf("Gives: %4d-%02d-%02d, %02d:%02d:%02d\n",
		r.year, r.mon, r.mday, r.hour, r.min, r.sec);
	
	return(0);
} // */

// *************************************************************************
// 
// *************************************************************************
static int test_keys()
{
	// test keyboard input
	term tty;
	int t, i;
	term_keypress k;
	char bn[64], be[64];
	
	term_open(&tty);
	
	for (t = 0; t < 10; t++) {
		term_getkey(&tty, &k);
		key_name(bn, 64, &k);
		key_escape(be, 64, &k);
		
		printf("%zd: ", k.n);
		
		for (i = 0; i < TERM_KEY_STR_LENGTH; i++) {
			if (i < (int)k.n) {
				printf(" %3d", (unsigned char)k.k[i]);
			} else {
				printf(" %3s", "");
			}
		}
		printf("   %-16s \"%s\"\n", bn, be);
	}
	
	term_close(&tty);
	return(0);
}

// *************************************************************************
// 
// *************************************************************************
int main(int argc, char **argv)
{
	int need_cfg, t, r;
	char *cfg_dir, *data_dir, *home;
	char *cfg_path, *cal_path, *cc;
	config cfg;
	
	cfg_path = NULL;
	cal_path = NULL;
	
	for (t = 1; t < argc; t++) {
		if ((!strcmp(argv[t], "-f"))
		||  (!strcmp(argv[t], "--config"))) {
			if (t + 1 >= argc) {
				DEBUG("-f: config path expected\n");
				return(-1);
			}
			if (cfg_path) {
				DEBUG("-f: config already specified"
					" ('%s', '%s')\n",
					cfg_path, argv[t + 1]);
				return(-1);
			}
			cfg_path = xprintf("%s", argv[++t]);
		} else if ((!strcmp(argv[t], "-c"))
		|| (!strcmp(argv[t], "--calendar"))) {
			if (t + 1 >= argc) {
				DEBUG("-c: calendar path expected\n");
				return(-1);
			}
			if (cal_path) {
				DEBUG("-c: calendar already specified"
					" ('%s', '%s')\n",
					cal_path, argv[t + 1]);
				return(-1);
			}
			cal_path = xprintf("%s", argv[++t]);
		} else if ((!strcmp(argv[t], "-v"))
		||  (!strcmp(argv[t], "--version"))) {
			PRINT("%s", VERSION_MSG);
			return(0);
		} else if ((!strcmp(argv[t], "-h"))
		||  (!strcmp(argv[t], "--help"))) {
			PRINT("%s", HELP_MSG);
			return(0);
		} else if (!strcmp(argv[t], "--test-keys")) {
			return(test_keys());
		} else {
			DEBUG("%s", HELP_MSG);
			return(-1);
		}
	}
	
	// env dirs for config path
	cfg_dir = getenv("XDG_CONFIG_HOME");
	data_dir = getenv("XDG_DATA_HOME");

	home = getenv("HOME");
	if (!home) {
		DEBUG("Warning: no HOME environment variable.\n");
	}

	// config
	config_init(&cfg);
	
	if (cfg_path) {
		need_cfg = 1;
	} else if (cfg_dir) {
		cfg_path = xprintf("%s/now.conf", cfg_dir);
		need_cfg = 0;
	} else if (home) {
		cfg_path = xprintf("%s/.config/now.conf", home);
		need_cfg = 0;
	} else {
		cfg_path = xprintf("%s", "now.conf");
		need_cfg = 0;
	}
	
	if (config_parse(&cfg, cfg_path)) {
		if (need_cfg)
			return(-1);
	}
	
	cc = config_get(&cfg, "calendar-path", NULL);
	
	if (cal_path)  {
		// use specified path
	} else if (cc) {
		// use path from configuration file
		cal_path = xprintf("%s", cc);
	} else if (data_dir) {
		// use default path in XDG data dir
		cal_path = xprintf("%s/now/calendar.ics", data_dir);
	} else if (home) {
		// use default path in home dir
		cal_path = xprintf("%s/.local/share/now/calendar.ics", home);
	} else {
		// use current directory
		cal_path = xprintf("calendar.ics");
	}
	
	r = ui_run(&cfg, cal_path);
	
	xfree(cfg_path);
	xfree(cal_path);
	
	config_clear(&cfg);
	
	return(r);
}


	/*
	// test iCal I/O
	event_list l;
	
	evlist_init(&l);
	evlist_ical_load(&l, "output.ics");
	evlist_ical_save(&l, "output2.ics");
	return(0);*/

	
	/*
	// test UTF-8
	char *s = "ȮéàËabcd";
	
	printf("%s: %zd - %zd\n", s, strlen(s), utf8_strlen(s));
	return(0);*/

