/*
    This file is part of "now", a simple calendar tool
    Copyright (C) 2012-2019  Laurent Poirrier

    "now" is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include "memory.h"

// ***************************************************************************
// 
// ***************************************************************************
void *dxmalloc(size_t n, const char *what,
	const char *func, const char *file, int line)
{
	void *r;
	
	r = malloc(n);
	
	if (!r) {
		if (n == 0)
			return(NULL);
	
		fprintf(stderr, "%s:%d: %s(): malloc(%s): "
			"failed allocating %zd bytes\n",
			file, line, func, what, n);
		exit(1);
	}
	
	return(r);
}


// ***************************************************************************
// 
// ***************************************************************************
void *dxrealloc(void *p, size_t n, const char *pname, const char *what,
	const char *func, const char *file, int line)
{
	void *r;
	
	r = realloc(p, n);
	
	if (!r) {
		if (n == 0)
			return(NULL);

		fprintf(stderr, "%s:%d: %s(): realloc(%s, %s): "
			"failed allocating %zd bytes\n",
			file, line, func, pname, what, n);
		exit(1);
	}
	
	return(r);
}


// ***************************************************************************
// 
// ***************************************************************************
void dxfree(void *p, const char *pname,
	const char *func, const char *file, int line)
{
	if (!p) {
		fprintf(stderr, "%s:%d: %s(): free(%s): "
			"freeing null pointer\n",
			file, line, func, pname);
		exit(1);
	}
	
	free(p);
}

// ***************************************************************************
// 
// ***************************************************************************
typedef struct tb_tag_t {
	void *ptr;
	size_t size;
	size_t space;
	struct tb_tag_t *head, *tail;
	struct tb_tag_t *parent, *prev, *next;
} tb_tag;

// ***************************************************************************
// 
// ***************************************************************************
static tb_tag *tb_get_tag(void *p)
{
	return(*((tb_tag **)(p - sizeof(tb_tag *))));
}

static void tb_set_tag(void *p, tb_tag *t)
{
	*((tb_tag **)(p - sizeof(tb_tag *))) = t;
}

// ***************************************************************************
// 
// ***************************************************************************
void *tb_sub_malloc(void *master, size_t n)
{
	tb_tag *t;
	void *p;
	
	p = xmalloc(sizeof(tb_tag *) + n) + sizeof(tb_tag *);
	t = xmalloc(sizeof(tb_tag));
	
	tb_set_tag(p, t);
	
	t->ptr = p;
	t->size = n;
	t->space = n;
	
	t->head = t->tail = NULL;
	t->parent = t->prev = t->next = NULL;
	
	if (master) {
		t->parent = tb_get_tag(master);
		t->prev = t->parent->tail;
		
		if (t->prev)
			t->prev->next = t;
		
		t->parent->tail = t;
	}
	
	return(p);
}

// ***************************************************************************
// 
// ***************************************************************************
void *tb_realloc(void *p, size_t n)
{
	tb_tag *t;
	size_t target;
	
	if (!p)
		return(tb_sub_malloc(NULL, n));
	
	t = tb_get_tag(p);
	
	if (t->space >= n) {
		t->size = n;
		return(p);
	}
	
	if (n > t->space * 2)
		target = n;
	else
		target = t->space * 2;
	
	p = xrealloc(p - sizeof(tb_tag *), sizeof(tb_tag *) + target)
		+ sizeof(tb_tag *);
	t->ptr = p;
	t->size = n;
	t->space = target;
	
	return(p);
}


// ***************************************************************************
// 
// ***************************************************************************
void tb_free(void *p)
{
	tb_tag *t;

	if (!p)
		return;
	
	t = tb_get_tag(p);
	
	while (t->tail)
		tb_free(t->tail);
	
	if (t->parent) {
		if (t->prev)
			t->prev->next = t->next;
		else
			t->parent->head = t->next;
		
		if (t->next)
			t->next->prev = t->prev;
		else
			t->parent->tail = t->prev;
	}
	
	if (t->size)
		xfree(t->ptr - sizeof(tb_tag *));
	xfree(t);
}

// ***************************************************************************
// 
// ***************************************************************************
void *tb_malloc(size_t n)
{
	return(tb_sub_malloc(NULL, n));
}

void *tb_master()
{
	return(tb_malloc(0));
}

size_t tb_size(void *p)
{
	if (!p)
		return(0);
	
	return(tb_get_tag(p)->size);
}


size_t tb_space(void *p)
{
	if (!p)
		return(0);
	
	return(tb_get_tag(p)->space);
}


size_t tb_group_size(void *p)
{
	tb_tag *t, *u0, *u1;
	size_t r;
	
	if (!p)
		return(0);
	
	t = tb_get_tag(p);
	r = t->size;
	
	u0 = t->head;
	while (u0) {
		u1 = u0->next;
		
		r += tb_group_size(u0->ptr);
		
		u0 = u1;
	}
	
	return(r);
}

size_t tb_group_space(void *p)
{
	tb_tag *t, *u0, *u1;
	size_t r;
	
	if (!p)
		return(0);
	
	t = tb_get_tag(p);
	r = t->space;
	
	u0 = t->head;
	while (u0) {
		u1 = u0->next;
		
		r += tb_group_size(u0->ptr);
		
		u0 = u1;
	}
	
	return(r);
}


// **************************************************************************
// 
// **************************************************************************
char *xvprintf(const char *format, va_list ap)
{
	int len;
	char *dst;
	va_list aq;
	
	va_copy(aq, ap);

	len = vsnprintf(NULL, 0, format, aq);
	
	va_end(aq);

	dst = xmalloc(len + 1);

	vsnprintf(dst, len + 1, format, ap);

	return(dst);
}


// **************************************************************************
// 
// **************************************************************************
char *xprintf(char *format, ...)
{
	va_list ap;
	char *r;
	
	va_start(ap, format);
	r = xvprintf(format, ap);
	va_end(ap);

	return(r);
}


