/*
    This file is part of "now", a simple calendar tool
    Copyright (C) 2012-2019  Laurent Poirrier

    "now" is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include "memory.h"
#include "bio.h"
#include "screen.h"

// *************************************************************************
// 
// *************************************************************************
void screen_init(screen *s)
{
	s->width = 0;
	s->height = 0;
	
	s->data = NULL;
	s->flags = NULL;
	s->drow = NULL;
	s->frow = NULL;
}

void screen_clear(screen *s)
{
	if (s->drow)
		xfree(s->drow);
	if (s->frow)
		xfree(s->frow);
	if (s->data)
		xfree(s->data);
	if (s->flags)
		xfree(s->flags);
}


void screen_set(screen *s, screen_flags flags, utf8_code c)
{
	int x, y;
	
	for (y = 0; y < s->height; y++) {
		for (x = 0; x < s->width; x++) {
			s->drow[y][x] = c;
			s->frow[y][x] = flags;
		}
	}
}

void screen_resize(screen *s, int width, int height)
{
	int y;
	
	if ((width <= s->width) && (height <= s->height)) {
		s->width = width;
		s->height = height;
		screen_set(s, TERM_DEFAULTS | SCREEN_TRASH, SCREEN_BLANK);
		return;
	}
	
	s->width = width;
	s->height = height;

	s->data = xrealloc(s->data, width * height * sizeof(utf8_code));
	s->flags = xrealloc(s->flags, width * height * sizeof(screen_flags));
	s->drow = xrealloc(s->drow, height * sizeof(char *));
	s->frow = xrealloc(s->frow, height * sizeof(screen_flags *));
	
	for (y = 0; y < height; y++) {
		s->drow[y] = &(s->data[y * width]);
		s->frow[y] = &(s->flags[y * width]);
	}
	
	screen_set(s, TERM_DEFAULTS | SCREEN_TRASH, SCREEN_BLANK);
}

// *************************************************************************
// 
// *************************************************************************
void screen_termflags(term *t, screen_flags f1)
{
	// defaults
	term_font(t, TERM_DEFAULTS, 0);

	// color
	if (!(f1 & TERM_DEFAULTS)) {
		term_color(t, (f1 >> 40) & 0xffffff, (f1 >> 16) & 0xffffff);
	}
	
	// font
	term_font(t, f1 & (~TERM_DEFAULTS) & 0xff, 0);
}


// *************************************************************************
// 
// *************************************************************************
void screen_fdump(screen *s, FILE *f)
{
	char buf[SCREEN_WIDTH_MAX * UTF8_RENDER_MAX];
	int w, y, n;
	
	w = s->width;
	if (w > SCREEN_WIDTH_MAX)
		w = SCREEN_WIDTH_MAX;

	for (y = 0; y < s->height; y++) {
		n = utf8_render(buf, s->drow[y], w);
		fwrite(buf, n, 1, f);
		fwrite("\n", 1, 1, f);
	}
}

static void screen_ldump(screen *s, term *t, int x0, int y, int w,
	screen_flags *fl)
{
	char buf[SCREEN_WIDTH_MAX * UTF8_RENDER_MAX];
	int x1, l, n;
	
	if (w > SCREEN_WIDTH_MAX)
		w = SCREEN_WIDTH_MAX;
	
	x1 = x0 + w - 1;
	
	while (x0 <= x1) {
		l = 0;
		while ((x0 + l <= x1) && (s->frow[y][x0 + l] == *fl))
			l++;
		
		//screen_termflags(t,
		//	(s->frow[y][x0] & (~(0xf<<20)) & ~TERM_DEFAULTS)
		//	| ((random() & 0x7) << 20));
		
		n = utf8_render(buf, &s->drow[y][x0], l);

		if (term_write(t, buf, n))
			return;

		if ((x0 + l <= x1) && (s->frow[y][x0 + l] != *fl)) {
			screen_termflags(t, s->frow[y][x0 + l]);
			*fl = s->frow[y][x0 + l];
		}
		
		x0 += l;
	}
}

void screen_tdump(screen *s, term *t)
{
	int y, w;
	screen_flags fl;
	
	term_goto(t, 0, 0);
	term_font(t, TERM_DEFAULTS, 0);
	fl = 0;
	
	for (y = 0; y < s->height; y++) {
		w = (y == s->height - 1) ? s->width - 1 : s->width;
		screen_ldump(s, t, 0, y, w, &fl);
	}
}

void screen_tup(screen *s, term *t, screen *bkg)
{
	int w, h, x, x0, x1, y;
	screen_flags fl;

	w = s->width;
	h = s->height;
	if ((bkg->width != w) || (bkg->height != h))
		screen_resize(bkg, w, h);
	
	term_font(t, TERM_DEFAULTS, 0);
	fl = TERM_DEFAULTS;
	
	for (y = 0; y < h; y++) {
		x0 = 0;
		while ((x0 < w) && (s->drow[y][x0] == bkg->drow[y][x0])
		&& ((s->frow[y][x0] & (~SCREEN_TRASH)) == bkg->frow[y][x0]))
			x0++;
		
		x1 = (y == s->height - 1) ? w - 2 : w - 1;

		while ((x1 >= x0) && (s->drow[y][x1] == bkg->drow[y][x1])
		&& ((s->frow[y][x1] & (~SCREEN_TRASH)) == bkg->frow[y][x1]))
			x1--;

		if (x1 >= x0) {
			term_goto(t, x0, y);
		
			screen_ldump(s, t, x0, y, x1 - x0 + 1, &fl);
		
			for (x = x0; x <= x1; x++) {
				bkg->drow[y][x] = s->drow[y][x];
				bkg->frow[y][x] = s->frow[y][x]
					& (~SCREEN_TRASH);
			}
		}
	}
}


// *************************************************************************
// 
// *************************************************************************
void screen_multi(screen *s, int x, int y, int l,
	screen_flags flags, utf8_code c)
{
	int o;
	
	if ((y < 0) || (y >= s->height))
		return;
	
	if (x < 0) {
		l -= -x;
		x = 0;
	}
	
	if (x + l > s->width)
		l = s->width - x;
	
	for (o = 0; o < l; o++) {
		s->drow[y][x + o] = c;
		s->frow[y][x + o] = flags;
	}
}

void screen_at(screen *s, int x, int y, int l, screen_flags flags, char *str)
{
	int o, slen;
	
	if ((y < 0) || (y >= s->height))
		return;
	
	if (x < 0) {
		str += -x;
		l -= -x;
		x = 0;
	}
	
	if (x + l > s->width)
		l = s->width - x;
	
	if (l < 0)
		l = 0;
	
	slen = utf8_strlen(str);
	if (slen > l)
		slen = l;
	
	utf8_extract(&s->drow[y][x], str, l);

	for (o = slen; o < l; o++) {
		s->drow[y][x + o] = SCREEN_BLANK;
	}
	
	for (o = 0; o < l; o++) {
		s->frow[y][x + o] = flags;
	}
}

void screen_atf(screen *s, int x, int y, int l,
	screen_flags flags, char *fmt, ...)
{
	va_list ap;
	int sl;
	char *store;
	
	va_start(ap, fmt);
	sl = vsnprintf(NULL, 0, fmt, ap);
	va_end(ap);
	
	if (sl <= 0) {
		screen_at(s, x, y, l, flags, "");
		return;
	}
	
	store = xmalloc(sl + 1);
	
	va_start(ap, fmt);
	vsnprintf(store, sl + 1, fmt, ap);
	va_end(ap);
	
	screen_at(s, x, y, l, flags, store);
	
	xfree(store);
}

