/*
    This file is part of "now", a simple calendar tool
    Copyright (C) 2012-2019  Laurent Poirrier

    "now" is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "daytime.h"
#include "timeparse.h"

// **************************************************************************
// 
// **************************************************************************
void canon_time(char *s, time_t t)
{
	dt_time r;
	
	dt_localtime(t, &r);
	
	snprintf(s, CANON_TIME_SIZE, "%04d-%02d-%02d %02d:%02d",
		r.year, r.mon, r.mday, r.hour, r.min);
}

// **************************************************************************
// 
// **************************************************************************
void canon_duration(char *s, time_t start, time_t t)
{
	if (t - start > 24 * 3600) {
		canon_time(s, t);
		return;
	}
	
	if ((t - start) % 60) {
		snprintf(s, CANON_TIME_SIZE, "+ %02ld:%02ld:%02ld",
			(t - start) / 3600, ((t - start) % 3600) / 60,
			(t - start) % 60);
	} else {
		snprintf(s, CANON_TIME_SIZE, "+ %02ld:%02ld",
			(t - start) / 3600, ((t - start) % 3600) / 60);
	}
}


// **************************************************************************
// 
// **************************************************************************
static long int strtol2(char **sp)
{
	char *s, *e;
	long int v;
	
	s = *sp;
	while (*s == ' ')
		s++;
	
	v = strtol(s, &e, 10);
	
	while (*e == ' ')
		e++;
	
	*sp = e;
	return(v);
}


// *******************************************************
static time_t parse_mktime(dt_time *r)
{
	if ((r->year < 1900) || (r->mon < 1) || (r->mon >12)
	||  (r->mday < 1) || (r->mday > 31) || (r->hour < 0)
	||  (r->hour > 23) || (r->min < 0) || (r->min > 60))
		return(-1);
	return(dt_mktime(r));
}

// *******************************************************
time_t parse_time(time_t init, char *s)
{
	time_t t;
	dt_time r;
	int v, v1;
	
	dt_localtime(init, &r);

	// get year|month|hour
	v = strtol2(&s);
	
	if ((*s == '-') || (*s == '.')) {
		s++;
		
		// get month|day
		v1 = strtol2(&s);

		if ((*s == '-') || (*s == '.')) {
			s++;
			r.year = v;
			r.mon = v1;
			
			// get day
			r.mday = strtol2(&s);
		} else {
			r.mon = v;
			r.mday = v1;
		}

		if (*s == 0) {
			t = parse_mktime(&r);
			return(t);
		}
		
		// get hour
		v = strtol2(&s);
	}
	
	if (*s != ':') {
		return(-1);
	}
	
	s++;
	r.hour = v;
	
	// get min
	r.min = strtol2(&s);
	
	if (*s == 0) {
		t = parse_mktime(&r);
		return(t);
	}
	
	if (*s != ':') {
		return(-1);
	}
	
	// get sec
	r.sec = strtol2(&s);
	
	if (*s != 0) {
		return(-1);
	}
	
	t = parse_mktime(&r);
	return(t);
}

// **************************************************************************
// 
// **************************************************************************
time_t parse_duration(time_t start, char *s)
{
	time_t t;
	int v;
	int l, k[2];
	
	while (*s == ' ')
		s++;
	
	if (*s != '+') {
		t = parse_time(start, s);
		if (t < start) {
			return(-1);
		}
		return(t);
	}
	
	s++;
	while (*s == ' ')
		s++;
		
	t = start;
	l = 0;

	while (*s) {
		v = strtol2(&s);
	
		if (!strncmp(s, "days", 4)) {
			t += v * 24 * 3600;
			s += 4;
		} else if (!strncmp(s, "day", 3)) {
			t += v * 24 * 3600;
			s += 3;
		} else if (*s == 'd') {
			t += v * 24 * 3600;
			s += 1;
		} else if (!strncmp(s, "hours", 5)) {
			t += v * 3600;
			s += 5;
		} else if (!strncmp(s, "hour", 4)) {
			t += v * 3600;
			s += 4;
		} else if (*s == 'h') {
			t += v * 3600;
			s += 1;
		} else if (!strncmp(s, "minutes", 7)) {
			t += v * 60;
			s += 7;
		} else if (!strncmp(s, "minute", 6)) {
			t += v * 60;
			s += 6;
		} else if (!strncmp(s, "mins", 4)) {
			t += v * 60;
			s += 4;
		} else if (!strncmp(s, "min", 3)) {
			t += v * 60;
			s += 3;
		} else if (*s == 'm') {
			t += v * 60;
			s += 1;
		} else if (!strncmp(s, "seconds", 7)) {
			t += v;
			s += 7;
		} else if (!strncmp(s, "second", 6)) {
			t += v;
			s += 6;
		} else if (!strncmp(s, "secs", 4)) {
			t += v;
			s += 4;
		} else if (!strncmp(s, "sec", 3)) {
			t += v;
			s += 3;
		} else if (*s == 'm') {
			t += v;
			s += 1;
		} else if ((*s == ':') && (l < 2)) {
			s++;
			k[l] = v;
			l++;
		} else if (*s == 0) {
			if (l == 0) {
				t += v * 60;
			} else if (l == 1) {
				t += k[0] * 3600 + v * 60;
			} else if (l == 2) {
				t += k[0] * 3600 + k[1] * 60 + v;
			} else {
				return(-1);
			}
			l = 0;
		} else {
			return(-1);
		}

		while (*s == ' ')
			s++;
	}
	
	if (l) {
		return(-1);
	}
	return(t);
}


