/*
    This file is part of "now", a simple calendar tool
    Copyright (C) 2012-2019  Laurent Poirrier

    "now" is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include "memory.h"
#include "bio.h"
#include "terminal.h"
#include "screen.h"
#include "keys.h"
#include "theme.h"
#include "table.h"
#include "dateproc.h"
#include "event.h"
#include "ical.h"
#include "dialog.h"
#include "log.h"
#include "ui.h"

// ****************************************************************************
// 
// ****************************************************************************
static char *dayname[7] = { "Sunday", "Monday", "Tuesday", "Wednesday",
	"Thursday", "Friday", "Saturday" };

static char *monthname_short[13] = { "000", "Jan", "Feb", "Mar", "Apr",
	"May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

// ****************************************************************************
// 
// ****************************************************************************
static void paint_day(screen *s, theme *th,
	cell *c, time_t day, event_list *l,
	int *sl, int *soffs, int *slast, event_handle **sev)
{
	time_t today;
	event_list dl;
	event_handle *h;
	dt_time r;
	char c0, c1;
	screen_flags fl, flp;
	int i;
	
	today = dt_daymid(dt_now());

	dt_localtime(day, &r);
	
	if ((sl) && (*sl < 0)) {
		fl = th->comp[T_UI_DAYLINE_SEL];
		flp = th->comp[T_UI_DAYLINE_MARK_SEL];
	} else {
		fl = th->comp[T_UI_DAYLINE];
		flp = th->comp[T_UI_DAYLINE_MARK];
	}
	
	if (day != today)
		flp = fl;
	
	if (sl) {
		if (day == today) {
			c0 = '*';
			c1 = '*';
		} else {
			c0 = '[';
			c1 = ']';
		}
	} else {
		if (day == today) {
			c0 = '=';
			c1 = '=';
		} else {
			c0 = ' ';
			c1 = ' ';
		}
	}
	
	screen_atf(s, c->x, c->y, c->width - 1, fl,
		":%s %2d:", monthname_short[r.mon], r.mday);
	screen_atf(s, c->x, c->y, (c->width >= 2) ? 1 : 0, flp, "%c", c0);
	screen_atf(s, c->x + 7, c->y, (c->width >= 9) ? 1 : 0, flp, "%c", c1);
	
	evlist_init(&dl);
	evlist_select(&dl, l, dt_daystart(day), dt_dayend(day));
	evlist_slowsort(&dl);
	h = dl.head;
	
	if ((sl) && (soffs) && (slast) && (sev) && (*sl >= 0)) {
		if (*sl > dl.size)
			*sl = dl.size;
		if (*sl < 0)
			*sl = 0;
		
		if (*soffs + c->height - 2 < *sl)
			*soffs = *sl - (c->height - 2);
		if (*soffs > *sl)
			*soffs = *sl;
		
		i = *soffs;
		while ((h) && (i > 0)) {
			h = h->next;
			i--;
		}
		
		*slast = dl.size;
		*sev = NULL;
	} else {
		soffs = NULL;
		slast = NULL;
		sev = NULL;
	}
	
	for (i = 0; i < c->height - 1; i++) {
		fl = th->comp[T_UI_EVENT];
		flp = th->comp[T_UI_EVENT_MARK];
		
		if ((sl) && (*sl >= 0) && (*soffs + i == *sl)) {
			fl = th->comp[T_UI_EVENT_SEL];
			flp = th->comp[T_UI_EVENT_MARK_SEL];
			if (h)
				*sev = h->orig;
		}
		
		if (i == c->height - 2) {
			fl |= th->comp[T_UI_EVENT_LAST_MOD];
		}
		
		if (h) {
			if (h->ev->start < dt_daystart(day)) {
				screen_atf(s, c->x, c->y + 1 + i, c->width - 1,
				fl, "  ... %s",
				h->ev->data[EVENT_SUMMARY]);
			} else {
				dt_localtime(h->ev->start, &r);
			
				screen_atf(s, c->x, c->y + 1 + i, c->width - 1,
					fl, "%02d:%02d %s", r.hour, r.min,
					h->ev->data[EVENT_SUMMARY]);
			}

			if (c->width >= 7) {
				if ((i == 0) && (soffs) && (*soffs > 0))
					screen_atf(s, c->x + 5, c->y + 1 + i, 1,
						flp, "^");
				else if ((i == c->height - 2) && (h->next))
					screen_atf(s, c->x + 5, c->y + 1 + i, 1,
						flp, "+");
			}
		
			h = h->next;
		} else {
			screen_at(s, c->x, c->y + 1 + i, c->width - 1, fl, "");
		}
	}
	
	evlist_clear(&dl);
}

// ****************************************************************************
// 
// ****************************************************************************
void ui_ev_edit(term *tty, key_list *kl, theme *th, screen *scr, screen *bkg,
	log *act, event_list *l, event_handle *selev, time_t day)
{
	event *ev;
	int r;
	
	if (selev) {
		ev = selev->ev;
	} else {
		ev = event_new();
		ev->start = day;
		ev->stop = ev->start + 3600;
	}
	
	r = ui_event(tty, kl, th, scr, bkg, ev);
	
	if (selev) {
		if (r == 0) {
			log_appendf(act, "Mod: %4d-%02d-%02d %02d:%02d %s",
				dt_year(ev->start), dt_mon(ev->start),
				dt_mday(ev->start),
				dt_hour(ev->start), dt_min(ev->start),
				ev->data[EVENT_SUMMARY]);
		}
	} else {
		if (r == 0) {
			evlist_attach(l, ev);
			log_appendf(act, "New: %4d-%02d-%02d %02d:%02d %s",
				dt_year(ev->start), dt_mon(ev->start),
				dt_mday(ev->start),
				dt_hour(ev->start), dt_min(ev->start),
				ev->data[EVENT_SUMMARY]);
		} else {
			event_destroy(ev);
		}
	}
}

// ****************************************************************************
// 
// ****************************************************************************
void ui_ev_delete(term *tty, key_list *kl, theme *th, screen *scr, screen *bkg,
	log *act, event_list *l, event_handle *selev)
{
	int r;
	
	if (!selev)
		return;
	
	r = ui_selectn(tty, kl, th, scr, bkg,
		selev->ev->data[EVENT_SUMMARY], 2,
		"Delete event.",
		"Keep event.");
	
	if ((r == -1) || (r == 1))
		return;
	
	log_appendf(act, "Del: %4d-%02d-%02d %02d:%02d %s",
		dt_year(selev->ev->start), dt_mon(selev->ev->start),
		dt_mday(selev->ev->start),
		dt_hour(selev->ev->start), dt_min(selev->ev->start),
		selev->ev->data[EVENT_SUMMARY]);

	event_destroy(selev->ev);
	evlist_detach(l, selev);
}

// ****************************************************************************
// 
// ****************************************************************************
int ui_ev_save(term *tty, key_list *kl, theme *th, screen *scr, screen *bkg,
	log *act, event_list *l, char *cal_path)
{
	char *xstr;
	
	if (b_create_dir_path(cal_path)) {
		xstr = xprintf("Could not create directory for %s: %s",
			cal_path, strerror(errno));
		ui_selectn(tty, kl, th, scr, bkg, xstr, 1,
			"Continue without saving.");
		xfree(xstr);
		return(-1);
	}
	
	if (evlist_ical_save(l, cal_path)) {
		xstr = xprintf("Could not save calendar (%s: %s)",
			cal_path, strerror(errno));
		ui_selectn(tty, kl, th, scr, bkg, xstr, 1,
			"Continue without saving.");
		xfree(xstr);
		return(-1);
	}
	
	log_crop(act, 5);
	return(0);
}

// ****************************************************************************
// 
// ****************************************************************************
int ui_ev_exit(term *tty, key_list *kl, theme *th, screen *scr, screen *bkg,
	log *act, event_list *l, char *cal_path)
{
	int r;
	
	if (act->n > 5) {
		log_modif(act, 0, "Save modifications and exit.");
		log_modif(act, 1, "Save modifications and continue editing.");
		log_modif(act, 2, "Discard modifications and exit.");
		
		while (1) {
			r = ui_select(tty, kl, th, scr, bkg,
				"Unsaved modifications to the calendar:",
				act->n, act->lines);
		
			if ((r == 0) || (r == 1)) {
				// save
				if (ui_ev_save(tty, kl, th, scr, bkg, act, l,
						cal_path))
					r = -1;
			}
			
			if ((r < 0) || (r == 1)) {
				// continue editing
				return(-1);
			}
			
			if ((r == 0) || (r == 2)) {
				// exit
				return(0);
			}
		}
	} else {
		r = ui_selectn(tty, kl, th, scr, bkg,
			"Exit (no modifications to calendar):", 2,
			"Yes, exit.",
			"No, continue editing.");
		
		if (r == 1)
			return(-1);
		else
			return(0);
	}
}

// ****************************************************************************
// 
// ****************************************************************************
#define UI_LOCK_EXIT	-1
#define UI_LOCK_NOF_RO	-2
#define UI_LOCK_NOF_RW	-3
#define UI_LOCK_RO	-4
#define UI_LOCK_RW	-5
int ui_ev_lock(term *tty, key_list *kl, theme *th, screen *scr, screen *bkg,
	log *act, event_list *l, char *cal_path)
{
	int fd, r;
	char *xstr;

	fd = -1;
	
	while (1) {
		fd = open(cal_path, O_RDONLY);
		
		if (fd >= 0)
			break;
		
		xstr = xprintf("Could not read calendar (%s: %s)",
			cal_path, strerror(errno));
		r = ui_selectn(tty, kl, th, scr, bkg, xstr, 3,
			"Create empty calendar.",
			"Start with empty calendar, create file when saving.",
			"Exit now.");
		xfree(xstr);
		
		if (r == 0) {
			ui_ev_save(tty, kl, th, scr, bkg, act, l, cal_path);
		} else if (r == 1) {
			break;
		} else {
			return(UI_LOCK_EXIT);
		}
	}

	if (fd >= 0) {
		r = flock(fd, LOCK_EX | LOCK_NB);
		
		if (r == 0)
			return(fd);
		
		close(fd);
	}

	if (fd < 0) {
		xstr = xprintf("An other instance could be running.");
	} else if (errno == EWOULDBLOCK) {
		xstr = xprintf("An other instance is running.");
	} else {
		xstr = xprintf("Could not acquire lock on calendar "
			"(%s: %s).", cal_path, strerror(errno));
	}

	r = ui_selectn(tty, kl, th, scr, bkg, xstr, 3,
		"Keep going read-only.",
		"Keep going read-write, dangerous!",
		"Exit now.");

	xfree(xstr);
	
	if (r == 0)
		return((fd >= 0) ? UI_LOCK_RO : UI_LOCK_NOF_RO);
	if (r == 1)
		return((fd >= 0) ? UI_LOCK_RW : UI_LOCK_NOF_RW);

	return(UI_LOCK_EXIT);
}

void ui_ev_unlock(int fd)
{
	if (fd >= 0) {
		flock(fd, LOCK_UN);
		close(fd);
	}
}

// ****************************************************************************
// 
// ****************************************************************************
int ui_run(config *cfg, char *cal_path)
{
	// config
	int first_wday;

	// output related stuff
	term tty;
	screen cur, bkg;
	table tab;
	int level, x, y, sx, sy, sl, soffs, slast;
	theme th;
	
	// input related stuff
	key_list kl;
	term_keypress k;
	
	// event related stuff
	event_list l;
	log act;

	// working variables
	time_t start, day, today, selday;
	event_handle *selev;
	int fd, rw;

	// Init log
	log_init(&act);
	log_appendf(&act, "");
	log_appendf(&act, "");
	log_appendf(&act, "");
	log_appendf(&act, "");
	log_appendf(&act, "----------------------------------------");

	// Init terminal input/output
	if (term_open(&tty))
		return(1);

	key_init(&kl);
	key_config(&kl, cfg);
	
	theme_init(&th);
	theme_config(&th, cfg);
	
	if (config_get_bool(cfg, "alt-screen", 1))
		term_altscreen(&tty, 1);
	
	x = config_get_choice(cfg, "terminal-colors", 0,
		4, "default", "none", "no", "0", "1",
		"16", "256", "16M", "16m");
	
	if ((x >= 1) && (x <= 4))
		tty.palette = COLOR_NONE;
	else if (x == 5)
		tty.palette = COLOR_16;
	else if (x == 6)
		tty.palette = COLOR_256;
	else if ((x >= 7) &&  (x <= 8))
		tty.palette = COLOR_16M;
		
	screen_init(&cur);
	screen_init(&bkg);
	table_init(&tab, 7, 6);
	level = 0;
	
	if (term_getsize(&tty))
		level = -1;

	// Init calendar
	first_wday = config_get_weekday(cfg, "first-weekday", 1);

	evlist_init(&l);
	
	// Acquire lock
	fd = ui_ev_lock(&tty, &kl, &th, &cur, &bkg, &act, &l, cal_path);
	
	if (fd == UI_LOCK_EXIT) {
		rw = 0;
		level = -1;
	} else if ((fd == UI_LOCK_NOF_RO) || (fd == UI_LOCK_RO)) {
		rw = 0;
	} else {
		rw = 1;
	}
	
	// Load
	if ((fd >= 0) || (fd == UI_LOCK_RO) || (fd == UI_LOCK_RW)) {
		if (evlist_ical_load(&l, cal_path))
			level = -1;
	}
	
	// Init main loop
	today = dt_daymid(dt_now());
	start = dt_weekstart(dt_monthstart(today), first_wday);
	sy = (today + (3600 * 12) - start) / (3600 * 24);
	sx = sy % tab.width;
	sy = sy / tab.width;
	sl = -1;
	soffs = slast = 0;

	// Main loop
	while (level >= 0) {
		if ((level == 0) || (term_signaled(&tty))) {
			if (term_getsize(&tty))
				return(1);

			screen_resize(&cur, tty.width, tty.height);
			screen_resize(&bkg, tty.width, tty.height);
			
			screen_set(&cur, th.comp[T_UI_BKG], SCREEN_BLANK);
			
			table_fit(&tab, 0, 1, tty.width, tty.height - 2);
		}
		
		level = 1;

		for (x = 0; x < tab.width; x++) {
			screen_atf(&cur,
				tab.elem[0][x].x, 0,
				tab.elem[0][x].width - 1,
				th.comp[T_UI_TOPLINE],
				"%s", dayname[(x + first_wday) % 7]);
		}
		
		day = start;
		selday = 0;
		selev = NULL;
		for (y = 0; y < tab.height; y++) {
			for (x = 0; x < tab.width; x++) {
				if ((x == sx) && (y == sy)) {
					selday = day;
					paint_day(&cur, &th,
						&tab.elem[y][x], day,
						&l, &sl, &soffs, &slast,
						&selev);
				} else {
					paint_day(&cur, &th,
						&tab.elem[y][x], day,
						&l, NULL, NULL, NULL, NULL);
				}
				
				day = dt_nextday(day, 1);
			}
		}
		
		screen_atf(&cur, 0, tty.height - 1, tty.width - 1,
			th.comp[T_UI_BOTTOMLINE], "%4d-%02d-%02d ",
			dt_year(selday), dt_mon(selday), dt_mday(selday));
		
		if (selev) {
			if (selev->ev->start < selday - 12 * 3600) {
				screen_atf(&cur, 11, tty.height - 1,
					tty.width - 12,
					th.comp[T_UI_BOTTOMLINE],
					"| %s (%4d.%02d.%02d %02d:%02d - "
					"%4d.%02d.%02d %02d:%02d)",
					selev->ev->data[EVENT_SUMMARY],
					dt_year(selev->ev->start),
					dt_mon(selev->ev->start),
					dt_mday(selev->ev->start),
					dt_hour(selev->ev->start),
					dt_min(selev->ev->start),
					dt_year(selev->ev->stop),
					dt_mon(selev->ev->stop),
					dt_mday(selev->ev->stop),
					dt_hour(selev->ev->stop),
					dt_min(selev->ev->stop));
			} else {
				screen_atf(&cur, 11, tty.height - 1,
					tty.width - 12,
					th.comp[T_UI_BOTTOMLINE],
					"| %02d:%02d %s (%d h %d min)",
					dt_hour(selev->ev->start),
					dt_min(selev->ev->start),
					selev->ev->data[EVENT_SUMMARY],
					(selev->ev->stop - selev->ev->start)
						/ 3600,
					((selev->ev->stop - selev->ev->start)
						% 3600) / 60);
			}
		}
		
		screen_tup(&cur, &tty, &bkg);
		term_goto(&tty, tty.width - 1, tty.height - 1);
		
		term_getkey(&tty, &k);
		
		if (sl < 0) {
			if (key_action(&kl, &k,A_CAL_ENTER)) {
				sl = 0;
			} else if (key_action(&kl, &k, A_CAL_WEEK_PREV)) {
				sy--;
			} else if (key_action(&kl, &k, A_CAL_WEEK_NEXT)) {
				sy++;
			} else if (key_action(&kl, &k, A_CAL_DAY_PREV)) {
				sx--;
			} else if (key_action(&kl, &k, A_CAL_DAY_NEXT)) {
				sx++;
			} else if (key_action(&kl, &k, A_CAL_EXIT)) {
				if (!ui_ev_exit(&tty, &kl, &th, &cur, &bkg,
						&act, &l, cal_path))
					level = -1;
				else
					level = 0;
			} else if (key_action(&kl, &k, A_CAL_SHIFT_M1)) {
				start = dt_nextweek(start, -1);
				sy++;
				if (sy > tab.height - 1)
					sy = tab.height - 1;
			} else if (key_action(&kl, &k, A_CAL_SHIFT_P1)) {
				start = dt_nextweek(start, 1);
				sy--;
				if (sy < 0)
					sy = 0;
			} else if (key_action(&kl, &k, A_CAL_SHIFT_M4)) {
				start = dt_nextweek(start, -4);
			} else if (key_action(&kl, &k, A_CAL_SHIFT_P4)) {
				start = dt_nextweek(start, 4);
			} else if (key_action(&kl, &k, A_CAL_WEEK_START)) {
				sx = 0;
			} else if (key_action(&kl, &k, A_CAL_WEEK_END)) {
				sx = tab.width - 1;
			} else if (key_action(&kl, &k, A_CAL_WEEK_START_M1)) {
				first_wday = (first_wday + 6) % 7;
				start = dt_nextday(start, -1);
				sx++;
			} else if (key_action(&kl, &k, A_CAL_WEEK_START_P1)) {
				first_wday = (first_wday + 1) % 7;
				start = dt_nextday(start, 1);
				sx--;
			}
		} else {
			if (key_action(&kl, &k, A_DAY_LEAVE)) {
				sl = -1;
			} else if (key_action(&kl, &k, A_DAY_EV_PREV)) {
				sl--;
				if (sl < 0) {
					sy--;
					sl = 1 << 24;
				}
			} else if (key_action(&kl, &k, A_DAY_EV_NEXT)) {
				sl++;
				if (sl > slast) {
					sy++;
					sl = 0;
				}
			} else if (key_action(&kl, &k, A_DAY_DAY_PREV)) {
				sl = 0;
				sx--;
			} else if (key_action(&kl, &k, A_DAY_DAY_NEXT)) {
				sl = 0;
				sx++;
			} else if (key_action(&kl, &k, A_DAY_EV_FIRST)) {
				sl = 0;
			} else if (key_action(&kl, &k, A_DAY_EV_LAST)) {
				sl = 1 << 24;
			} else if (key_action(&kl, &k, A_DAY_SHIFT_M1)) {
				start = dt_nextweek(start, -1);
				sy++;
				if (sy > tab.height - 1) {
					sy = tab.height - 1;
					sl = 0;
				}
			} else if (key_action(&kl, &k, A_DAY_SHIFT_P1)) {
				start = dt_nextweek(start, 1);
				sy--;
				if (sy < 0) {
					sy = 0;
					sl = 0;
				}
			} else if (key_action(&kl, &k, A_DAY_SHIFT_M4)) {
				start = dt_nextweek(start, -4);
				sl = 0;
			} else if (key_action(&kl, &k, A_DAY_SHIFT_P4)) {
				start = dt_nextweek(start, 4);
				sl = 0;
			} else if ((rw) && (key_action(&kl, &k, A_DAY_EDIT))) {
				ui_ev_edit(&tty, &kl, &th, &cur, &bkg,
					&act, &l, selev, selday);
				level = 0;
			} else if ((rw) && (selev)
			&& (key_action(&kl, &k, A_DAY_DELETE))) {
				ui_ev_delete(&tty, &kl, &th, &cur, &bkg,
					&act, &l, selev);
				level = 0;
			}
		}
		if (sx < 0) {
			sx = tab.width - 1;
			sy--;
		}
		if (sx > tab.width - 1) {
			sx = 0;
			sy++;
		}
		
		if (sy < 0) {
			sy = 0;
			start = dt_nextweek(start, -1);
		}
		
		if (sy > tab.height - 1) {
			sy = tab.height - 1;
			start = dt_nextweek(start, 1);
		}
	}
	
	// Release lock
	ui_ev_unlock(fd);
	
	// Cleanup screen
	term_goto(&tty, tty.width - 1, tty.height - 1);
	term_font(&tty, TERM_DEFAULTS, 0);
	term_write(&tty, "\n", 1);

	if (config_get_bool(cfg, "alt-screen", 1))
		term_altscreen(&tty, 0);

	screen_clear(&bkg);
	screen_clear(&cur);
	table_clear(&tab);
	key_clear(&kl);
	term_close(&tty);
	
	// Cleanup memory
	evlist_destroy_all(&l);
	evlist_clear(&l);
	log_clear(&act);

	return(0);
}

