/*
    This file is part of "now", a simple calendar tool
    Copyright (C) 2012-2019  Laurent Poirrier

    "now" is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "utf8.h"

// Ken Thompson > RFC

size_t utf8_bytes(char *ss)
{
	unsigned char *s = (unsigned char *)ss;
	unsigned char c;
	size_t expect, b;
	
	c = *s;
	
	if (c == 0x00)
		return(0);
	if ((c & 0x80) == 0x00)
		return(1);
	
	if ((c & 0xe0) == 0xc0)
		expect = 2;
	else if ((c & 0xf0) == 0xe0)
		expect = 3;
	else if ((c & 0xf8) == 0xf0)
		expect = 4;
	else if ((c & 0xfc) == 0xf8)
		expect = 5;
	else
		expect = 6;
	
	for (b = 1; b < expect; b++) {
		if ((s[b] & 0xc0) != 0x80) {
			return(b);
		}
	}
	
	return(expect);
}

size_t utf8_full(char *ss, size_t bmax)
{
	unsigned char *s = (unsigned char *)ss;
	unsigned char c;
	size_t b;
	
	c = *s;
	
	if ((c & 0x80) == 0x00)
		return(1);
	if ((c & 0xe0) == 0xc0)
		return(2);
	if ((c & 0xf0) == 0xe0)
		return(3);
	if ((c & 0xf8) == 0xf0)
		return(4);
	if ((c & 0xfc) == 0xf8)
		return(5);
	if ((c & 0xfe) == 0xfc)
		return(6);
	
	if (bmax > 6)
		bmax = 6;
	
	for (b = 0; b < bmax; b++) {
		if ((s[b] & 0xc0) != 0x80) {
			return(b);
		}
	}
	
	return((bmax == 6) ? 6 : (bmax + 1));
}

utf8_code utf8_encode(char *ss)
{
	unsigned char *s = (unsigned char *)ss;
	unsigned char c;
	size_t expect, b;
	utf8_code r;
	
	c = *s;
	
	if ((c & 0x80) == 0x00)
		return(c);
	
	if ((c & 0xe0) == 0xc0) {
		expect = 2;
		r = c & 0x1f;
	} else if ((c & 0xf0) == 0xe0) {
		expect = 3;
		r = c & 0x0f;
	} else if ((c & 0xf8) == 0xf0) {
		expect = 4;
		r = c & 0x07;
	} else if ((c & 0xfc) == 0xf8) {
		expect = 5;
		r = c & 0x03;
	} else {
		expect = 6;
		r = c & 0x01;
	}
	
	for (b = 1; b < expect; b++) {
		c = s[b];
		if ((c & 0xc0) != 0x80) {
			return(0xfffd);
		}
		r = (r << 6) | (c & 0x3f);
	}
	
	return(r);
}


size_t utf8_strlen(char *s)
{
	size_t b;
	
	b = 0;
	while (*s) {
		s += utf8_bytes(s);
		b++;
	}
	
	return(b);
}

size_t utf8_extract(utf8_code *dst, char *s, size_t lmax)
{
	size_t b;
	
	b = 0;
	while ((*s) && (b < lmax)) {
		dst[b++] = utf8_encode(s);
		s += utf8_bytes(s);
	}
	
	return(b);
}

size_t utf8_render(char *dsts, utf8_code *src, size_t l)
{
	unsigned char *dst = (unsigned char *)dsts;
	size_t o, b;
	utf8_code c;
	
	b = 0;
	for (o = 0; o < l; o++) {
		c = src[o];
		
		if ((c & 0xffffff80) == 0) {
			dst[b++] = c;
		} else if ((c & 0xfffff800) == 0) {
			dst[b++] = 0xc0 | (c >> 6);
			dst[b++] = 0x80 | (c & 0x3f);
		} else if ((c & 0xffff0000) == 0) {
			dst[b++] = 0xe0 | (c >> 12);
			dst[b++] = 0x80 | ((c >> 6) & 0x3f);
			dst[b++] = 0x80 | (c & 0x3f);
		} else if ((c & 0xffe00000) == 0) {
			dst[b++] = 0xf0 | (c >> 18);
			dst[b++] = 0x80 | ((c >> 12) & 0x3f);
			dst[b++] = 0x80 | ((c >> 6) & 0x3f);
			dst[b++] = 0x80 | (c & 0x3f);
		} else if ((c & 0xfc000000) == 0) {
			dst[b++] = 0xf8 | (c >> 24);
			dst[b++] = 0x80 | ((c >> 18) & 0x3f);
			dst[b++] = 0x80 | ((c >> 12) & 0x3f);
			dst[b++] = 0x80 | ((c >> 6) & 0x3f);
			dst[b++] = 0x80 | (c & 0x3f);
		} else {
			dst[b++] = 0xfc | ((c >> 30) & 0x1);
			dst[b++] = 0x80 | ((c >> 24) & 0x3f);
			dst[b++] = 0x80 | ((c >> 18) & 0x3f);
			dst[b++] = 0x80 | ((c >> 12) & 0x3f);
			dst[b++] = 0x80 | ((c >> 6) & 0x3f);
			dst[b++] = 0x80 | (c & 0x3f);
		}
	}
	
	return(b);
}


